@@ -123,6 +123,7 @@ spthread_t;
 #define STARTTLS_CMD        "STARTTLS"
 #define BDAT_CMD            "BDAT"
 #define XCLIENT_CMD         "XCLIENT"
+#define XFORWARD_CMD        "XFORWARD"
 
 #define DATA_END_SIG        "." CRLF
 
@@ -161,7 +162,6 @@ spthread_t;
 #define CFG_TIMEOUT         "TimeOut"
 #define CFG_OUTADDR         "OutAddress"
 #define CFG_LISTENADDR      "Listen"
-#define CFG_HEADER      	"Header"
 #define CFG_TRANSPARENT     "TransparentProxy"
 #define CFG_DIRECTORY       "TempDirectory"
 #define CFG_KEEPALIVES      "KeepAlives"
@@ -202,6 +202,7 @@ static int make_connections(spctx_t* ctx, int client);
 static int read_server_response(spctx_t* ctx);
 static int parse_config_file(const char* configfile);
 static char* parse_address(char* line);
+static char* parse_xforward(char* line, const char* part);
 static const char* get_successful_rsp(const char* line, int* cont);
 static void do_server_noop(spctx_t* ctx);
 
@@ -551,7 +552,7 @@ static void connection_loop(int sock)
         /* Check to make sure we have a thread */
         if(fd != -1)
         {
-            sp_messagex(NULL, LOG_ERR, "too many connections open (max %d). sent 554 response", g_state.max_threads);
+            sp_messagex(NULL, LOG_ERR, "too many connections open (max %d). sent busy response", g_state.max_threads);
             write(fd, SMTP_STARTBUSY, KL(SMTP_STARTBUSY));
             shutdown(fd, SHUT_RDWR);
             close(fd);
@@ -578,6 +579,7 @@ static void connection_loop(int sock)
                 close(fd);
             }
           
+            sp_messagex(NULL, LOG_DEBUG, "cleaning up completed thread");
             pthread_join(threads[i].tid, NULL);
             threads[i].tid = 0;
         }
@@ -648,6 +650,12 @@ static void cleanup_context(spctx_t* ctx)
         ctx->sender = NULL;
     }
     
+    if(ctx->xforwardaddr)
+    {
+        free(ctx->xforwardaddr);
+        ctx->xforwardaddr = NULL;
+    }
+    
     ctx->logline[0] = 0;
 }
 
@@ -883,7 +891,7 @@ static int smtp_passthru(spctx_t* ctx)
                     RETURN(-1);
                     
                 /* 
-                 * Now go into avcheck mode. This also handles the eventual
+                 * Now go into scan mode. This also handles the eventual
                  * sending of the data to the server, making the av check
                  * transparent
                  */
@@ -1069,7 +1077,7 @@ static int smtp_passthru(spctx_t* ctx)
                     }
                 }
 
-                /* MAIL FROM */
+                /* MAIL FROM (that the server accepted) */
                 if((r = check_first_word(C_LINE, FROM_CMD, KL(FROM_CMD), SMTP_DELIMS)) > 0)
                 {
                     t = parse_address(C_LINE + r);
@@ -1081,7 +1089,7 @@ static int smtp_passthru(spctx_t* ctx)
                         strcpy(ctx->sender, t);
                 }
                 
-                /* RCPT TO */
+                /* RCPT TO (that the server accepted) */
                 else if((r = check_first_word(C_LINE, TO_CMD, KL(TO_CMD), SMTP_DELIMS)) > 0)
                 {
                     t = parse_address(C_LINE + r);
@@ -1101,6 +1109,21 @@ static int smtp_passthru(spctx_t* ctx)
                         strcat(ctx->recipients, t);
                     }
                 }
+                            
+                /*
+                 * If the client sends an XFORWARD, and the server accepted it, 
+                 * we store address for use in our forked process environment 
+                 * variables (see sp_setup_forked). 
+                 */
+                else if(is_first_word(C_LINE, XFORWARD_CMD, KL(XFORWARD_CMD)))
+                {
+                    if((t = parse_xforward (C_LINE + KL(XFORWARD_CMD), "ADDR"))) 
+                    {
+                        ctx->xforwardaddr = (char*)reallocf(ctx->xforwardaddr, strlen(t) + 1);
+                        if(ctx->xforwardaddr)
+                            strcpy(ctx->xforwardaddr, t);
+                    }
+                }
                 
                 /* RSET */
                 else if(is_first_word(C_LINE, RSET_CMD, KL(RSET_CMD)))
@@ -1154,6 +1177,40 @@ static char* parse_address(char* line)
     return trim_end(line);
 }
 
+static char* parse_xforward(char* line, const char* part)
+{
+    char* t;
+    char* e;
+
+    t = strcasestr(line, part);
+    if(!t)
+        return NULL;
+
+    /* equals sign and surrounding */
+    t = trim_start(t + strlen(part));
+    if(*t != '=')
+        return NULL;
+    t = trim_start(t + 1);
+    if(!*t)
+        return NULL;
+    
+    /* Find the end of the thingy */
+    if(*t == '[')
+    {
+        t++;
+        e = strchr(t, ']');
+    }
+    else
+    {
+        e = t + strcspn(t, " \t");
+    }
+    
+    if(!e)
+        return NULL;
+    *e = 0;
+    return t;
+}
+
 static const char* get_successful_rsp(const char* line, int* cont)
 {
     /*
@@ -1487,13 +1544,14 @@ static int make_header(spctx_t* ctx, const char* format_str, char* header)
     return l >= MAX_HEADER_LENGTH ? MAX_HEADER_LENGTH - 1 : l; 
 }
 
-int sp_done_data(spctx_t* ctx)
+int sp_done_data(spctx_t* ctx, const char *headertmpl)
 {
     FILE* file = 0;
     int ret = 0;
     char *line;    
     char header[MAX_HEADER_LENGTH] = "";
     size_t header_len, line_len;
+    int header_prepend = 0;
     ssize_t rc;
 
     ASSERT(ctx->cachename[0]);  /* Must still be around */
@@ -1534,16 +1592,21 @@ int sp_done_data(spctx_t* ctx)
 
     sp_messagex(ctx, LOG_DEBUG, "sending from cache file: %s", ctx->cachename);
     
-    if(g_state.header)
-        header_len = make_header(ctx, g_state.header, header);
+    if(headertmpl)
+    {
+        header_len = make_header(ctx, headertmpl, header);
+        if(is_first_word(RCVD_HEADER, header, KL(RCVD_HEADER)))
+            header_prepend = 1;
+    }
+
 
     /* If we have to prepend the header, do it */
-    if(header[0] != '\0' && g_state.header_prepend)
+    if(header[0] != '\0' && header_prepend)
     {
-	    if(spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)header, header_len) == -1 ||
-	       spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)CRLF, KL(CRLF)) == -1)
-	        RETURN(-1);
-		header[0] = '\0';
+        if(spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)header, header_len) == -1 ||
+           spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)CRLF, KL(CRLF)) == -1)
+            RETURN(-1);
+        header[0] = '\0';
     }
 
     /* Transfer actual file data */    
@@ -1569,7 +1632,7 @@ int sp_done_data(spctx_t* ctx)
                 if(spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)header, header_len) == -1 ||
                    spio_write_data_raw(ctx, &(ctx->server), (unsigned char*)CRLF, KL(CRLF)) == -1)
                     RETURN(-1);
-                header[0] = '\0';				
+                header[0] = '\0';
             }
         }
         
@@ -1698,6 +1761,9 @@ void sp_setup_forked(spctx_t* ctx, int file)
     if(spio_valid(&(ctx->client)))
         setenv("CLIENT", ctx->client.peername, 1);
     
+    if(ctx->xforwardaddr)
+        setenv("REMOTE", ctx->xforwardaddr, 1);
+    
     if(spio_valid(&(ctx->server)))
         setenv("SERVER", ctx->server.peername, 1);
     
@@ -1710,13 +1776,12 @@ void sp_setup_forked(spctx_t* ctx, int file)
  */
  
 const char kMsgDelimiter[] = ": ";
-#define MAX_MSGLEN  256
+#define MAX_MSGLEN  1024
 
 static void vmessage(spctx_t* ctx, int level, int err, 
                      const char* msg, va_list ap)
 {
-    size_t len;
-    char* m;  
+    char buf[MAX_MSGLEN];
     int e = errno;
 
     if(g_state.daemonized)
@@ -1732,35 +1797,30 @@ static void vmessage(spctx_t* ctx, int level, int err,
        
     ASSERT(msg);
 
-    len = strlen(msg) + 20 + MAX_MSGLEN;
-    m = (char*)alloca(len);
-
-    if(m)
+    if(ctx)
+        snprintf(buf, MAX_MSGLEN, "%06X: %s%s", ctx->id, msg, err ? ": " : "");
+    else
+        snprintf(buf, MAX_MSGLEN, "%s%s", msg, err ? ": " : "");
+            
+    if(err)
     {
-        if(ctx)
-            snprintf(m, len, "%06X: %s%s", ctx->id, msg, err ? ": " : "");
-        else
-            snprintf(m, len, "%s%s", msg, err ? ": " : "");
+        /* strerror_r doesn't want to work for us for some reason
+        len = strlen(buf);
+        strerror_r(e, buf + len, MAX_MSGLEN - len); */
             
-        if(err)
-        {
-            /* strerror_r doesn't want to work for us for some reason
-            strerror_r(e, m + strlen(m), MAX_MSGLEN); */
-            
-            sp_lock();
-                strncat(m, strerror(e), len);
-            sp_unlock();
-        }
-            
-        m[len - 1] = 0;
-        msg = m;
-    }    
+        sp_lock();
+            strncat(buf, strerror(e), MAX_MSGLEN);
+        sp_unlock();
+    }
+
+    /* As a precaution */
+    buf[MAX_MSGLEN - 1] = 0;
   
     /* Either to syslog or stderr */
     if(g_state.daemonized)
-        vsyslog(level, msg, ap);
+        vsyslog(level, buf, ap);
     else
-        vwarnx(msg, ap);  
+        vwarnx(buf, ap);  
 }
 
 void sp_messagex(spctx_t* ctx, int level, const char* msg, ...)
@@ -1923,16 +1983,6 @@ int sp_parse_option(const char* name, const char* value)
         ret = 1;
     }
 
-    else if(strcasecmp(CFG_HEADER, name) == 0)
-    {
-        g_state.header = trim_start(value);
-	    if(strlen(g_state.header) == 0)
-	        g_state.header = NULL;
-	    else if(is_first_word(RCVD_HEADER, g_state.header, KL(RCVD_HEADER)))
-	        g_state.header_prepend = 1;
-	    ret = 1;
-    }
-    
     /* Always pass through to program */    
     if(cb_parse_option(name, value) == 1)
         ret = 1;
